// mtbw.h: S.W. Ellingson, Virginia Tech, 2010 Sep 23
// auxillary source code for mtbw.c
// See end of this file for history.

#define TBWH_N_STANDS_MAX 261 
#define TBWH_N_SAMPLES_MAX 36000000
#define TBWH_N_FRAMES_PER_STAND 30001  /* This is the maximum number of frames per polarization in a single trigger */
                                       /* = 12 000 000 samples /  400 samples/packet for 12-bit data */
                                       /* = 36 000 000 samples / 1200 samples/packet for  4-bit data */

#define TBWH_STATUS_OK 0
#define TBWH_STATUS_FILE_NOT_FOUND -1
#define TBWH_STATUS_DEC0DE5C_NOT_FOUND -2
#define TBWH_STATUS_NOT_TBW -3 /* not seeing the "TBW data" flag bit in TBW_ID */
#define TBWH_STATUS_4BIT_CHANGED -4 /* TBW_ID bit flag indicating bits/sample changed mid-file */
 
#define SHOW_SAMPLES 0

/* Storage of remembering time tags for every packet */
static long int ttt[TBWH_N_STANDS_MAX][TBWH_N_FRAMES_PER_STAND];

int tbw_file_info( char     *filename,  /* filename of TBW-generated file (via DR) to check */
                   int      *b4bit,     /* 1 means data is self-identified as 4-bit (as opposed to 12-bit) data */
                   long int *nFrames,   /* total number of frames seen (counting 0xDEC0DE5C's) */
                   int      *iStandMax, /* index (1..260) of highest-numbered stand seen in TBW_ID */ 
                   long     *time0,     /* value of first time tag seen */
                   int      bVerbose    /* =1 means "be verbose" */
                 ) {
/* Reads the specified filename and returns b4bit, nFrames, and iStandMax */ 
/* Also, the stand data is parsed and written binary to files tbw###.dat where ### is stand number */
  
  int eStatus = TBWH_STATUS_OK; /* Assessment of file (until we find otherwise) */                  

  FILE *fp;
  FILE *fpo[TBWH_N_STANDS_MAX];

  int b,b2,bb;
  int k;
  char outfilename[256];
  
  int bDEC0DE5C = 0; /* Have we ever seen the start-of-packet pattern? */
  int bStand[TBWH_N_STANDS_MAX]; /* =1 means we've seen this stand in the data */
  int iStand;
  int b4bit0 = -1; /* "-1" indicates we have not read the 4/12-bit flag bit of TBW_ID yet */ 
  long int iFrame[TBWH_N_STANDS_MAX];
  long int iFrameMax;

  long int l;
  long int ttt0 = -1;

  union {
    unsigned long int i;
    unsigned char b[8];
    } i8u;

  short int x0, y0;
  char c;
  char x0c,y0c;

  /* initialize variables */
  *nFrames = 0;
  *iStandMax = 0;
  for (k=0;k<TBWH_N_STANDS_MAX;k++) { 
    bStand[k]=0;
    iFrame[k]=0;
    for (l=0;l<TBWH_N_FRAMES_PER_STAND;l++) {
      ttt[k][l] = 0;
      }
    }
  iStand = 0;

  /* attempt to open file */
  if (!(fp = fopen(filename,"rb"))) {
    eStatus = TBWH_STATUS_FILE_NOT_FOUND;
    return eStatus; 
    }

  /* looping through packets */
  while ( ( (b = fgetc(fp)) != EOF ) && (eStatus==TBWH_STATUS_OK) ) {

    /* Read the 0xDEC0DE5C (4-byte) pattern */
    bDEC0DE5C = 1;
                   if (b!=0xDE) bDEC0DE5C=0;
    b = fgetc(fp); if (b!=0xC0) bDEC0DE5C=0;
    b = fgetc(fp); if (b!=0xDE) bDEC0DE5C=0;
    b = fgetc(fp); if (b!=0x5C) bDEC0DE5C=0;
   
    if (bDEC0DE5C) {

      (*nFrames)++; /* track total number of frames seen */
      if (bVerbose) printf("tbw_file_info: Saw 0xDEC0DE5C; nFrames is now %ld\n",*nFrames);

      /* Skip next 8 bytes (ID, FrameCount, SecondsCount) */
      for (k=0;k<8;k++) b = fgetc(fp);

      /* Next 2 bytes are TBW_ID */
      /* ... first byte */
      b = fgetc(fp); //printf("TBW_ID = 0x%02x",b);
      if (bVerbose) printf("tbw_file_info:   Byte 1 of TBW_ID is 0x%02x\n",b);

      if ( b & 0x80 ) { /* bit set identifying this as TBW data */ 

          *b4bit = ( b & 0x40 ) >> 6; 
          if (bVerbose) printf("tbw_file_info:   *b4bit = %d\n",*b4bit);

          if ( b4bit0 < 0 ) b4bit0 = *b4bit;  /* set this the first time we see it */
          if ( *b4bit == b4bit0 ) {           /* this hasn't changed mid-file... */

              /* second byte of TBW_ID contains stand ID */
              b2 = fgetc(fp); //printf("%02x\n",b);
              if (bVerbose) printf("tbw_file_info:   Byte 2 of TBW_ID is 0x%02x\n",b2);
              bb = ( b << 8 ) + ( b2 & 0xFF ); /* concatenating the two bytes */
              bb = bb & 0x3FFF; /* resetting 2 high order bits */ 
              if (bVerbose) printf("tbw_file_info:   Stand ID is %d\n",bb);
              if ( b > *iStandMax ) *iStandMax = bb; /* if this is the highest-numbered stand we've seen, remember that */ 
              if (!bStand[bb]) { /* if not done already, open an output file for this stand */
                bStand[bb] = 1;  /* remember that we've already opened an output file for this stand */
                sprintf(outfilename,"mtbw%03d.dat",bb);
                //printf("%d %s\n",b,outfilename);
                fpo[bb] = fopen(outfilename,"wb");
                }
              iStand = bb;       /* remember that this is where we are */
              iFrame[iStand]++; /* remember how many packets for this stand we've seen */

              /* Skip next 2 bytes (Unassigned) */
              for (k=0;k<2;k++) b = fgetc(fp);

              /* Next 8 bytes are TimeTag */
              i8u.b[7] = (char) fgetc(fp); //printf("TimeTag = 0x%02x",b);
              i8u.b[6] = (char) fgetc(fp); //printf("%02x",b);
              i8u.b[5] = (char) fgetc(fp); //printf("%02x",b);
              i8u.b[4] = (char) fgetc(fp); //printf("%02x",b);
              i8u.b[3] = (char) fgetc(fp); //printf("%02x",b);
              i8u.b[2] = (char) fgetc(fp); //printf("%02x",b);
              i8u.b[1] = (char) fgetc(fp); //printf("%02x",b);
              i8u.b[0] = (char) fgetc(fp); //printf("%02x\n",b);
              if (bVerbose) printf("tbw_file_info:   Time tag is %ld\n",i8u.i);
              if (ttt0==-1) { 
                ttt0 = i8u.i;
                *time0 = i8u.i;  
                } /* first time tag we've seen; normalize rest relative to this */
              ttt[iStand][iFrame[iStand]] = i8u.i - ttt0;

              //if (tt0==0) { tt0 = tt; } 
              //printf("TimeTag = %ld dt = %ld\n",tt,tt-tt0);

              /* X0, Y0 come in 3 byte groups */
              if (!(*b4bit)) { /* this is 12-bit data */
                  //  12 bit samples:      X0(11:4), [X0(3:0) Y0(11:8)], Y0(7:0)
                  for (k=0;k<400;k++) {
                    b = fgetc(fp); x0 = ( b << 4 );
                    b = fgetc(fp); x0 = x0 + ( ( b & 0x00F0 ) >> 4 ); y0 = ( b << 8 ) & 0x0F00;
                    b = fgetc(fp);                                    y0 = y0 + ( b & 0x00FF );   
                    x0 = x0 ^ 0x0800; /* flip high bit */        
                    y0 = y0 ^ 0x0800; /* flip high bit */ 
                    //printf("%d %d\n",x0,y0);
                    fwrite(&x0,sizeof(short int),1,fpo[iStand]);
                    fwrite(&y0,sizeof(short int),1,fpo[iStand]);
                    }
                } else { /* this is 4-bit data */
                  //  4 bit samples:       [X0(3:0) Y0(3:0)], [X1(3:0) Y1(3:0)], [X2(3:0) Y2(3:0)]
                  for (k=0;k<1200;k++) {
                    c = fgetc(fp); 
                    fwrite(&c,sizeof(char),1,fpo[iStand]);
                    if (SHOW_SAMPLES) {
                      x0c = ( c >> 4 ); 
                      y0c = c & 0x0F;
                      x0c = x0c ^ 0x8; /* flip high bit */        
                      y0c = y0c ^ 0x8; /* flip high bit */ 
                      x0 = (short int) 0x000F & x0c;
                      y0 = (short int) 0x000F & y0c;
                      printf("tbw_file_info:   k=%4d: %3d %3d 0x%02x 0x%02x\n",k,x0,y0,x0,y0);
                      }   
                    }
                } /* if (!(*b4bit)) */

            } else { /* b4bit changed mid-file */

              eStatus = TBWH_STATUS_4BIT_CHANGED;

            } /* if ( *b4bit == b4bit0 ) */

        } else { /* *Not* being identified as TBW data */

          printf("Not seeing TBW flag; *nFrames=%ld b=%x\n",*nFrames,b);
          eStatus = TBWH_STATUS_NOT_TBW;

        } /* if ( b & 0x80 ) */

      } else { 

        //printf("Lost 0xDEC0DE5C. *nFrames=%ld feof(fp)=%d\n",*nFrames,feof(fp)); 
        eStatus = TBWH_STATUS_DEC0DE5C_NOT_FOUND;

      } /* if (bDEC0DE5C) */

    } /* while ( (b = fgetc(fp)) != EOF ) && (eStatus==TBWH_STATUS_OK) */

  /* close files */
  fclose(fp);
  for (k=1;k<=(*iStandMax);k++) { fclose(fpo[k]); }

  /* Figure out the highest number of frames for any stand */
  iFrameMax = 0;
  for (k=1;k<=(*iStandMax);k++) {
    if (iFrame[k]>iFrameMax) iFrameMax=iFrame[k];
    }

  /* Now, write out time tags */
  fp = fopen("mtbw_tt.dat","w");
  for (l=1;l<=iFrameMax;l++) {
    for ( k=1; k<=(*iStandMax); k++ ) {
      fprintf(fp,"%12ld ",ttt[k][l]);
      }
    fprintf(fp,"\n");
    }
  fclose(fp);

  return eStatus;
  }


int tbw_sfile_import( int iStand,    /* look in file tbw###.dat where ###==iStand */
                      int iPol,      /* 1 means use x-pol, 2 means use y-pol */
                      int b4bit,     /* 1 means this is 4-bit data */
                      short int *tx, /* where the data goes */
                      long int *n    /* number of samples read */ 
                    ) {
  /* Reads the specified filename and returns data for the specified polarization in tx[] */
  
  char filename[256];
  FILE *fp;
  long int k=0;
  short int x12;
  short int y12;
  char c; 

  //printf("[XX] Now looking at iStand=%d, iPol=%d\n",iStand,iPol); 

  sprintf(filename,"mtbw%03d.dat",iStand);
  if (!(fp = fopen(filename,"rb"))) { return 0; } 

  if (!b4bit) { /* 12 bit data */
      while ( fread(&x12,sizeof(short int),1,fp) > 0 ) {
        fread(&y12,sizeof(short int),1,fp);
        k++;
        if (iPol==1) { tx[k-1] = x12; }
        if (iPol==2) { tx[k-1] = y12; }
        }
    } else { /* 4 bit data */
      while ( fread(&c,sizeof(char),1,fp) > 0 ) {
        k++;
        if (iPol==1) { tx[k-1] = (short int) ( 0x000F & (( c >> 4   ) ^ 0x08 ) ); }
        if (iPol==2) { tx[k-1] = (short int) ( 0x000F & (( c & 0x0F ) ^ 0x08 ) ); }   
        }
    }

  //printf("%d %ld\n",feof(fp),k);
  //printf("tx[%ld] = %hd, x12 = %hx\n",k-1,tx[k-1],x12);
  
  fclose(fp);
  *n = k;     /* remember number of samples */
  return 1;
  }

//==================================================================================
//=== HISTORY ======================================================================
//==================================================================================
// mtbw.h: S.W. Ellingson, Virginia Tech, 2010 Sep 23
//   .1: initial version
// mtbw.h: S.W. Ellingson, Virginia Tech, 2010 Sep 30
//   .1: added support for -v (verbose) option 
//       fixed TBW_ID->StandID bug (was assuming StandID was 1 byte before)

//==================================================================================
//=== BELOW THIS LINE IS SCRATCH ===================================================
//==================================================================================
